#!/usr/bin/perl -w
#
# moleRat.pl v1.0
# Last modified Octboer 24, 2006 11:26
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose:
#	Calculates rates of evolution along the branches of one or more (gene) trees with respect to a dated
#	reference tree, both for each tree individually and across the set of trees as a whole. Also identifies
#	branches and clades that are evolving significantly differently from the overall average or that have
#	changed their rate significantly with respect to an ancestral reference point (as determined using a
#	paired Student's t-test and a paired Fisher's sign test).
#
# Input:
#	1) A nexus- or phylip-formatted tree file containing a description of the reference (super)tree
#	2) A tab-delimited file (node number<tab>age) containing the ages of the nodes in the reference (super)tree
#	2) A nexus-formatted tree file containing the descriptions of one or more gene (sub)trees (with branch lengths)
#
# Output:
#	1) geneRates_moleRat.txt or geneWindowRates_moleRat.txt
#		A tab-delimited file detailing average (nested) clade rate of evolution for each input gene. For the
#		latter file (invoked with -w), partition-specific rates for a given gene are also compared
#		statistically to the overall rate for that gene using paired comparisons (across branches).
#	2) branchRates_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file detailing whether individual branches are evolving significantly fast or slow
#		across the set of gene trees as whole
#	3) branchMatrix_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file giving the branch-specific rate of evolution of each gene for each branch on
#		the supertree
#	4) branchShifts_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file indicating individual branches that show a significant difference in rate
#		compared to that of an ancestral branch
#	5) cladeRates_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file detailing whether clades (phylogenetically corrected rate of all descendant
#		branches) are evolving significantly faster or slower across the set of gene trees as whole
#	6) cladeMatrix_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file giving the clade-specific rate of evolution  of each gene for each clade on
#		the supertree
#	7) cladeShifts_moleRat.txt (on by default; suppress with -b)
#		A tab-delimited file indicating individual clades that show a significant difference in rate
#		compared to that of an ancestral clade
#	8) <gene>_moleRat.txt (optional; switch on with -i)
#		A tab-delimited file for each (gene) tree listing branch- and clade-specific rates
#
# Usage: moleRat.pl -d<filename> -g<filename> -t<filename> [-b] [-c] [-i] [-l] [-n<number>] [-p] [-r] [-s] [-w] [-v] [-h]
#   options: -b = only calculate gene-specific rates (suppress branch- and clade-specific rates)
#			 -c = incorporate clade sizes when calculating phylogenetically-corrected clade-specific rates
#			 -d<filename> = tab-delimited file with node ages for reference (super)tree of form node number<tab>node age
#			 -g<filename> = tree file containing gene (sub)trees in NEXUS format
#			 -i = output results for each individual gene (sub)tree to file <gene>_moleRat.txt
#			 -l = do not ln-transform rates prior to calculating paired differences
#			 -n<number> = do not process branches in gene trees of this length or shorter (default = 0.0001; set to a negative
#						  number to include all branches)
#			 -p = add columns to output files to facilitate graphing of significant vs. not significant results
#				  (based on uncorrected P-values)
#			 -r = assume trees to be rooted (= branch lengths from root are meaningful) unless otherwise
#				  specified (default = unrooted)
#			 -s<number> = maximum nodal depth of ancestral clade to use for determining shifts in branch- and
#						  clade-specific rates (default = 3; 0 = unlimited)
#            -t<filename> = tree file containing reference (super)tree in NEXUS or phylip formats
#			 -w = perform window analysis on genes in which rates from different partitions for that gene (where tree
#				  names are given as gene_windowX or gene_codonX, where X = 1, 2, 3, ...) are compared to that across
#				  all partitions (where tree must proceed others and be named gene_all)
#			 -v = verbose output
#            -h = print this message and quit

use strict;
use constant PI => 3.1415926536;
use POSIX;

# Set user-input defaults and associated parameters
	# Supertree parsing
		my (%STnodeList, $numSTclade, @STtaxa, %taxonST, @STnodeList, %STnestedClades, %STdaughterClades);
		my %nodeAge;
	
	# Gene tree parsing
		my (%mismatchFlag, %skipTree);
		my %treeRoot;
		my $defaultRoot = 0;
		my %geneRate;
		my %branchMatrix;
			my (%branchZscore, %branchNumZ, %branchAvgZ, %branchVarZ, %branchseZ, %branchpZ);
			my (%branchNegCount, %branchZeroCount, %branchPlusCount, %branchpSign);
		my %cladeMatrix;
			my (%cladeZscore, %cladeNumZ, %cladeAvgZ, %cladeVarZ, %cladeseZ, %cladepZ);
			my (%cladeNegCount, %cladeZeroCount, %cladePlusCount, %cladepSign);
		my (%windowMatrix, %windowRefTree);
		my @GTnames;
		my $depthThreshold = 3;
		
	# I/O variables
		my ($superFile, $geneFile, $dateFile);
		my $logTransform = 1;	# Log transform values (using ln) prior to calculating paired differences
		my $graphPrint = 0;	# Add extra columns geared toward creating graphs
		my $genePrint = 0;	# Output equivalent statistics for each individual gene
		my $branchTests = 1;	# Restrict analysis to deriving gene rates
		my $windowAnalysis = 0;
		my $sizeWeighting = 0;	# Calculate phylogenetic corrections for clade-specific rates as non-weighted average of clade sizes
		my $lengthThreshold = 0.0001;
		my @fileNames = qw(branchRates_moleRat.txt branchShifts_moleRat.txt cladeRates_moleRat.txt cladeShifts_moleRat.txt);
		
	# Partition metric variables
		# For partition metric
			my $compType = "pairwise";	# Options are "pairwise" (all), and "ref" (all vs. first tree)
			my $weighted_dS = 0;	# Flag as to whether or not to calculate dS weighted by nodal support; assumes latter range from 0-100 (e.g., as for bootstrap frequencies)
			my (%dSlist, %norm_dSlist, %pruneSize);
			my %cladeWeight;
			my $rootingStatus;
			
		# Pruning variables
			my @treeSpecies;
			my $keepSupport = 1;
			my $pruneType = "common";	# Options are "common" (prune to taxon on all trees), "directed" (prune specific taxa), and "both"
				my %pruneTaxon;

	# Tree variables
		my (%treeName, %treeDescription, %rootingStatus);
		my %prunedTree;
		my $unrootedFlag = 0;
		my $defaultRooting = "[&U]";	# Default is for trees to be unrooted
		my $SupportFlag = 0;
		my @allTaxa;
		my $numTaxa;
		my $treeNum = 0;

	# Global variables
		my (@AllTrees, %treeNames);
		my $block;
			my $treeBlock = 0;
		my %factorial;
		my $skip_dS = 0;
		my $verbose = 0;
		my $debug = 0;
		my $version = "1.0";

# Process switches
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] eq "-b")
			{
			$branchTests = 0;
			$genePrint = 1;
			}
		elsif ($ARGV[$i] eq "-c")
			{
			$sizeWeighting = 1;
			}
		elsif ($ARGV[$i] =~ /^-d(.+)/)
			{
			$dateFile = $1;
			die "ERROR: Date file $dateFile does not exist.\n" unless (-e $dateFile);
			}
		elsif ($ARGV[$i] =~ /^-g(.+)/)
			{
			$geneFile = $1;
			die "ERROR: Treefile $geneFile does not exist.\n" unless (-e $geneFile);
			}
		elsif ($ARGV[$i] eq "-i")
			{
			$genePrint = 1;
			}
		elsif ($ARGV[$i] eq "-l")
			{
			$logTransform = 1;
			}
		elsif ($ARGV[$i] =~ /^-n(\-?\d*\.?\d+)/)
			{
			$lengthThreshold = $1;
			}
		elsif ($ARGV[$i] eq "-p")
			{
			$graphPrint = 1;
			}
		elsif ($ARGV[$i] eq "-r")
			{
			$defaultRoot = 1;
			}
		elsif ($ARGV[$i] =~ /^-s(\d+)/)
			{
			$depthThreshold = $1;
			}
		elsif ($ARGV[$i] =~ /^-t(.+)/)
			{
			$superFile = $1;
			die "ERROR: Treefile $superFile does not exist.\n" unless (-e $superFile);
			}
		elsif ($ARGV[$i] eq "-w")
			{
			$windowAnalysis = 1;
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-z")
			{
			$skip_dS = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: moleRat.pl -d<filename> -g<filename> -t<filename> [-b] [-c] [-i] [-l] [-n<number>] [-p] [-r] [-s] [-w] [-v] [-h]\n";
			print "Version: $version\n";
			print "Options: -b = only calculate gene-specific rates (suppress branch- and clade-specific rates)\n";
			print "         -c = incorporate clade sizes when calculating phylogenetically-corrected clade-specific rates\n";
			print "         -d<filename> = tab-delimited file with node ages for reference (super)tree of form node number<tab>node age\n";
			print "         -g<filename> = tree file containing gene trees in NEXUS format\n";
			print "         -i = output results for each individual gene (sub)tree to file <gene>_moleRat.txt\n";
			print "         -l = do not ln-transform rates prior to calculating paired differences\n";
			print "         -n<number> = do not process branches in gene trees of this length or shorter (default = 0.0001; set to a negative\n";
			print "                      number to include all branches)\n";
			print "         -p = add columns to output files to facilitate graphing of significant vs. not significant results\n";
			print "              (based on uncorrected P-values)\n";
			print "         -r = assume trees to be rooted (= branch lengths from root are meaningful) unless otherwise\n";
			print "              specified (default = unrooted)\n";
			print "         -s<number> = maximum nodal depth of ancestral clade to use for determining shifts in branch- and\n";
			print "                      clade-specific rates (default = 3; 0 = unlimited)\n";
			print "         -t<filename> = tree file containing reference (super)tree in NEXUS or phylip formats\n";
			print "         -w = perform window analysis on genes in which rates from different partitions for that gene (where tree\n";
			print "              names are given as gene_windowX or gene_codonX, where X = 1, 2, 3, ...) are compared to that across\n";
			print "              all partitions (where tree must proceed others and be named gene_all)\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			print "\nCitation: Bininda-Emonds, O.R.P. Submitted. Fast genes and slow clades: comparative rates of molecular evolution in mammals. Submitted to Genome Biology.\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: moleRat.pl -d<filename> -g<filename> -t<filename> [-b] [-c] [-i] [-l] [-n<number>] [-p] [-r] [-s] [-w] [-v] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}

	if (not defined $superFile or not defined $dateFile)
		{
		print "ERROR: Must specify name of file(s) containing:\n";
		print "\tsupertree\n" if (not defined $superFile);
		print "\tdivergence dates for supertree\n" if (not defined $dateFile);
		print "\tgene tree(s)\n" if (not defined $geneFile);
		exit(1);
		}

my $rateZero = time;

# Read in supertree, parse, and store information
	print "Reading in and parsing supertree description from file $superFile ...\n";
	my $superZero = time; 
	$block = 0;
	treeRead($superFile);
		my $superTree = $treeDescription{"supertree"} = $treeDescription{"treeID_1"};
			$superTree =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove any branch length information
			$numSTclade = ($superTree =~ tr/(//);
				die "ERROR: Supertree does not have enough informative nodes ($numSTclade clades)\n" if ($numSTclade < 2);
			print "\nSupertree description:\t$superTree\n" if ($verbose);

	# Parse supertree to get clade memberships and sizes
		my ($cladeID, %ParentClade, %CladeDescrip, %CladeSupport, %STcladeSize, %open, %start, %stop, %present);
		my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
		my $OldCurrent;
		while ($position < length($superTree))
			{
			my $offset = 1;
			if (substr($superTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeNum++;
					$cladeID = "STnode" . $cladeNum;
					push @STnodeList, $cladeID;
					$branchNegCount{$cladeID} = $branchZeroCount{$cladeID} = $branchPlusCount{$cladeID} = 0;
					$cladeNegCount{$cladeID} = $cladeZeroCount{$cladeID} = $cladePlusCount{$cladeID} = 0;
				$ParentClade{$cladeID} = $CurrentClade;
					$ParentClade{$cladeID} = "STnode0" if ($cladeID eq "STnode1");
					push @ { $STdaughterClades{$CurrentClade} }, $cladeID;
				$CurrentClade = $cladeID;
				$open{$CurrentClade} = 1;
					for (my $i = 1; $i <= $cladeNum; $i++)
						{
						my $clade = "STnode" . $i;
						if ($open{$clade})
							{
							push @ { $STnestedClades{$clade} }, $CurrentClade;
							}
						}
				$start{$CurrentClade} = $position;
				}
			elsif (substr($superTree, $position, 1) eq ")")	#Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				$stop{$OldCurrent} = $position;
				$CladeDescrip{$OldCurrent} = substr($superTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
				}
			elsif (substr($superTree, $position, 1) eq ",")
				{
				# Do nothing
				}
			else	# Determine membership of clade
				{
				until (substr($superTree, $position+$offset, 1) eq "(" or substr($superTree, $position+$offset, 1) eq ")" or substr($superTree, $position+$offset, 1) eq "," or substr($superTree, $position+$offset, 1) eq ":")
					{
					$offset++;
					}
				my $taxon = substr($superTree, $position, $offset);
					push @STtaxa, $taxon;
					$taxonST{$taxon} = 1;
					push @STnodeList, $taxon;
						$branchNegCount{$taxon} = $branchZeroCount{$taxon} = $branchPlusCount{$taxon} = 0;
						$nodeAge{$taxon} = 0;	# Terminal taxa have age of 0 (i.e., the present)
					$ParentClade{$taxon} = $CurrentClade;
					push @ { $STdaughterClades{$CurrentClade} }, $taxon;
				for (my $i = 1; $i <= $cladeNum; $i++)	# Assign all open clades to taxon and increment size of open clades
					{
					my $clade = "STnode" . $i;
					if ($open{$clade})
						{
						push @{ $STnodeList{$taxon} }, $clade;
						push @ { $STnestedClades{$clade} }, $taxon;
						$STcladeSize{$clade}++;
						}
					}
				$position += $offset-1;
				}
			$position++;
			}
	
		@STtaxa = sort @STtaxa;
	
		if ($debug)
			{
			print "\nTaxon membership and size of supertree clades:\n";
				for (my $i = 1; $i <= $numSTclade; $i++)
					{
					my $clade = "STnode" . $i;
					print "\t$clade ($STcladeSize{$clade}): $CladeDescrip{$clade}\n" if ($verbose);
					print "\t\tNested clades:\n";
						print "\t\t\t$_\n" foreach (@ { $STnestedClades{$clade} });
					} 
			print "\nSupertree clade membership of taxa:\n";
				foreach my $taxon (@STtaxa)
					{
					print "\t$taxon:";
					foreach my $node (@{ $STnodeList{$taxon} })
						{
						print "\t$node";
						}
					print "\n";
					}
			}
	printf "\tTime taken: %s seconds\n", time - $superZero;

# Read in ages of supertree from tab-delimited file
	print "Reading in nodal ages from file $dateFile ...\n";
	my $dateZero = time; 
	setLineBreak($dateFile);
	open (DATE, "<$dateFile") or die "Cannot open file containing nodal ages, $dateFile\n";
		while (<DATE>)
			{
			chomp;
			next unless ($_);
			next if ($_ =~ /^\s*\D/);
				print "\t$_\n" if ($debug);
			
			my @dateLine = split (/\s+/, $_);
				my $node = "STnode" . $dateLine[0];
				$nodeAge{$node} = (pop(@dateLine) * 10**6);
				print "\t$node: $nodeAge{$node}\n" if ($verbose);
			}
	printf "\tTime taken: %s seconds\n", time - $dateZero;

# Read in gene trees and store
	print "Reading in gene trees ...\n";
	my $allGeneZero = time;
	$block = 1;
	my @geneTrees = treeRead($geneFile);
	
	if ($windowAnalysis)
		{
		open (TREE, ">geneWindowRates_moleRat.txt") or die "Cannot write average rates of evolution for each tree to file geneWindowRates_moleRat.txt\n";
		print TREE "Tree\tAverage\tSE\tn\tPaired difference\tSE\tPaired-t\tPaired-t p-value\tplusCount\tnegCount\tProportion\tPaired-sign p-value\n";
		}
	else
		{
		open (TREE, ">geneRates_moleRat.txt") or die "Cannot write average rates of evolution for each tree to file geneRates_moleRat.txt\n";
		print TREE "Tree\tAverage\tSE\tn\n";
		}

	# Parse gene tree descriptions and get relative branch lengths
		foreach my $treeID (@geneTrees)	# Get structure of each tree
			{
			$treeDescription{$treeID} =~ s/:\d+\.?\d*e?\-?\d*$//i;	# Get rid of trailing branch length, which is uninformative and only causes problems when parsing branch lengths
				my $geneTree = $treeDescription{$treeID};
			my ($numGTrate, $varGTrate, $seGTrate);	# Overall rate variables for gene
			
			my $geneZero = time;
			my (%ParentClade, %daughterClade);
			my ($cladeID, %CladeDescrip, %open, %start, %stop, %GTcladeSize);
			my ($taxon, %geneNodeMember, @mismatchList);
			my (%CladeSupport, %nodeDepth, %cladeDepth, %cladeHeight);
				# Set variables for root
					$nodeDepth{GTnode0} = 0;
					$cladeDepth{GTnode0} = 0;
					$CladeSupport{GTnode1} = 0;			
			my %gene2ST;
			my @GTnodeList;
			
			print "\tProcessing tree $treeName{$treeID} ...\n";
				if ($verbose)
					{
					print "\t$geneTree\n";
					print "\t\tTree is ";
						print "un" if ($rootingStatus{$treeID} eq "[&U]");
						print "rooted\n";
					}
				if ($windowAnalysis)
					{
					if ($treeName{$treeID} =~ /_all$/)
						{
						$windowRefTree{$treeName{$treeID}} = 1;
						}
					else
						{
						my $rootName;
							($rootName = $treeName{$treeID}) =~ s/_[a-z]+\d+$/_all/;
						if (not defined $windowRefTree{$rootName})
							{
							print "\tERROR: Reference tree for window analysis not present; skipping tree ...\n";
							next;
							}
						}
					}
			# Do general check on tree structure
				print "\t\tChecking basic structure ...\n";
				# Check for mismatched / misnamed taxa
					my (%taxonCount, @GTrealTaxa, %specimenList);
					my $comparisonST = $superTree;
						$treeDescription{"comparisonST"} = $comparisonST;
					my $treeTaxa = $geneTree;
						$treeTaxa =~ s/:\d+\.?\d*e?\-?\d*//g;
						$treeTaxa =~ s/\(//g;
						$treeTaxa =~ s/\)//g;
						foreach my $taxonCheck (split(/,/, $treeTaxa))
							{
							my $rootTaxon;
								($rootTaxon = $taxonCheck) =~ s/_\d+$//;
								push @ { $specimenList{$rootTaxon} }, $taxonCheck;
							$taxonCount{$rootTaxon}++;
								push @GTrealTaxa, $rootTaxon if ($taxonCount{$rootTaxon} == 1);
							unless ($taxonST{$rootTaxon})	# List any taxa not present on supertree
								{
								push @mismatchList, $rootTaxon;
								}
							}
						if (@mismatchList)
							{
							$skipTree{$treeID} = 1;
							open (MISMATCH, ">>mismatch.txt") or die "Cannot append mismatched taxa to file mismatch.txt.\n";
								@mismatchList = sort (@mismatchList);
								print "\t\t\tERROR: Tree contains taxa not on supertree; list written to file mismatch.txt\n";
								print MISMATCH "$treeName{$treeID}:";
								foreach my $booboo (@mismatchList)
									{
									print MISMATCH "\t$booboo";
									}
								print MISMATCH "\n";
							close MISMATCH;
							}
						foreach my $taxon (@GTrealTaxa)
							{
							next unless ($taxonCount{$taxon} > 1);
							my $newTaxon = $taxon . "_1:";
							$treeDescription{$treeID} =~ s/$taxon\:/$newTaxon/;
							# Create new substitute clade for supertree
								my $multTaxon = "(";
								for (my $i = 1; $i <= $taxonCount{$taxon}; $i++)
									{
									$multTaxon .= $taxon . "_$i,";
									}
								$multTaxon =~ s/,$/)/;
							$comparisonST =~ s/$taxon/$multTaxon/;
								$treeDescription{"comparisonST"} = $comparisonST;
							}
				# Check tree size
					my $totalClade = ($geneTree =~ tr/(//);
						if ($totalClade < 2)
							{
							print "\t\t\tERROR: Tree has only one informative node\n";
							$skipTree{$treeID} = 1;
							}
				# Check for branch lengths
					unless ($geneTree =~ m/:\d+/)
						{
						print "\t\t\tERROR: Tree does not contain any branch lengths\n";
						$skipTree{$treeID} = 1;
						}
				# Check that gene tree and supertree are topologically identical
					print "\t\t\tTree identity check skipped ...\n" if ($skip_dS or $skipTree{$treeID});
					unless ($skipTree{$treeID} or $skip_dS)	# Expensive operation; only do if tree not already flagged
						{
						my $dS = partitionMetric($treeID, "comparisonST");
							if ($dS > 0)
								{
								print "\t\t\tERROR: Gene tree is not topologically identical with supertree (dS = $dS); skipping\n";
								$skipTree{$treeID} = 1;
								}
						}

				next if ($skipTree{$treeID});

			print "\t\tParsing structure ...\n";	# Tree passes: parse it!
				push @GTnames, $treeName{$treeID};
				
				# Correct any misnumbered specimens for a root taxon
					foreach my $rootTaxon (@GTrealTaxa)
						{
						my @sortList = sort(@ { $specimenList{$rootTaxon} });
						
						my $specimenCount = 0;
						foreach my $specimen (@sortList)
							{
							$specimenCount++;
							my $newName = $rootTaxon;
								$newName .= "_$specimenCount" if ($taxonCount{$rootTaxon} > 1);
							if ($specimen ne $newName)
								{
								$treeDescription{$treeID} =~ s/$specimen\:/$newName\:/;
								print "\t\t\tWARNING: $specimen renamed as $newName\n";
								}
							}
						}
					$geneTree = $treeDescription{$treeID};	# Rederive gene tree in case of any changes for multiple specimens

			my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
			my $OldCurrent;
			while ($position < length($geneTree))
				{
				my $offset = 1;
				if (substr($geneTree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$cladeNum++;
						$cladeID = "GTnode" . $cladeNum;
						push @GTnodeList, $cladeID;
					$ParentClade{$cladeID} = $CurrentClade;
						$ParentClade{$cladeID} = "GTnode0" if ($cladeID eq "GTnode1");
					push @ {$daughterClade{$CurrentClade} }, $cladeID;	# Assign new clade as daughter clade to parent clade
					$CurrentClade = $cladeID;
					$nodeDepth{$CurrentClade} = $nodeDepth{$ParentClade{$CurrentClade}} + 1;
					$open{$CurrentClade} = 1;
					$start{$CurrentClade} = $position;
					}
				elsif (substr($geneTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					$stop{$OldCurrent} = $position;

					# Get clade description
						$CladeDescrip{$OldCurrent} = substr($geneTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
						$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
						
						$GTcladeSize{$OldCurrent} = ($CladeDescrip{$OldCurrent} =~tr/\,//);	# Get size of clade
						
					print "\t\t\tClade ID: $OldCurrent\n" if ($debug);
					print "\t\t\t\t$CladeDescrip{$OldCurrent}\n" if ($debug);
					
					# Determine corresponding supertree clade
						my %count;
						print "\t\t\t\tTaxon membership:\n" if ($debug);
						foreach my $geneTaxon (@{ $geneNodeMember{$OldCurrent} })	# Check each species of node in gene tree
							{
							print "\t\t\t\t\t$geneTaxon\n" if ($debug);
							foreach my $STnode (@{ $STnodeList{$geneTaxon} })	# Find what nodes that species belongs to in supertree
								{
								$count{$STnode}++;
								}
							}
						# Find supertree node of smallest size for which all species in gene tree are present; this is the least inclusive node on supertree containing all taxa from the node on the gene tree
							print "\t\t\t\tFinding equivalent ST node:\n" if ($debug);
							if ($gene2ST{$OldCurrent})	# Clade is multiple specimens of a single species; no equivalent
								{
								print "\t\t\t\t\tnone\n" if ($debug);
								}
							else
								{
								my $maxSize = scalar(@STtaxa) + 1;	# Initialize counter to one larger than number of taxa on ST
								for (my $STnodeNum = 1; $STnodeNum <= $numSTclade; $STnodeNum++)
									{
									my $STnode = "STnode" . $STnodeNum;
									# Quick checks to avoid extra processing
										next unless ($count{$STnode});
										next if (scalar(@{ $geneNodeMember{$OldCurrent} }) > $STcladeSize{$STnode});	# Gene tree clade is more inclusive than ST clade
										next if ($STcladeSize{$STnode} > $maxSize);
									print "\t\t\t\t\tST node ID: $STnode ($STcladeSize{$STnode})\n" if ($debug);
									if ($count{$STnode} == scalar(@{ $geneNodeMember{$OldCurrent} }) and $STcladeSize{$STnode} < $maxSize)
		
										{
										$gene2ST{$OldCurrent} = $STnode;
										$maxSize = $STcladeSize{$STnode};
										print "\t\t\t\t\t\tCurrent equivalent\n" if ($debug);
										}
									}
								}
					# Determine absolute and relative heights from tips
						my $sumHeight = 0;
						print "\t\t\t\tDaughter clades:\n" if ($debug);
						foreach my $daughter (@ {$daughterClade{$OldCurrent} })
							{
							print "\t\t\t\t\t$daughter: $CladeSupport{$daughter}\n" if ($debug);
							$sumHeight += $CladeSupport{$daughter};							
							$sumHeight += $cladeHeight{$daughter};							
							}
						print "\t\t\t\t\tSum height: $sumHeight\n" if ($debug);
						$cladeHeight{$OldCurrent} = sprintf "%.5f", $sumHeight / scalar(@ {$daughterClade{$OldCurrent} });
							print "\t\t\t\t\tClade height: $cladeHeight{$OldCurrent}\n" if ($debug);
					}
				elsif (substr($geneTree, $position, 1) eq ":")	# Store branch lengths for node or taxon
					{
					until (substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq "," or ($position+$offset) > length($geneTree))
						{
						$offset++;
						}
					my $branchLength = substr($geneTree, $position+1, $offset-1);
						print "\t\t\t\tBranch length: $branchLength\n" if ($debug);
						if (substr($geneTree, $position-1, 1) eq ")")	# Branch length pertains to a clade
							{
							$CladeSupport{$OldCurrent} = substr($geneTree, $position+1, $offset-1);
							}
						else	# Branch length pertains to a terminal taxon
							{
							$CladeSupport{$taxon} = $branchLength;
							}
					$position += $offset-1;
					}
				elsif (substr($geneTree, $position, 1) eq ",")
					{
					# Do nothing
					}
				else	# Determine membership of clade
					{
					until (substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq "," or substr($geneTree, $position+$offset, 1) eq ":")
						{
						$offset++;
						}
					$taxon = substr($geneTree, $position, $offset);
						push @ { $daughterClade{$CurrentClade}}, $taxon;
						push @GTnodeList, $taxon;
						$ParentClade{$taxon} = $CurrentClade;
						$nodeDepth{$taxon} = $nodeDepth{$CurrentClade} + 1;
						$cladeHeight{$taxon} = 0;
					print "\t\t\tTaxon: $taxon\n" if ($debug);
					
					# Process taxon name according to whether it is a multiple specimen or not
						if (not defined $taxonST{$taxon})	# Taxon of form Genus_species_x (if ($taxon =~ m/_\d+$/)), designating repeated specimen of Genus_species; derive real taxon name
							{
							$gene2ST{$CurrentClade} = $gene2ST{$taxon} = "none";	# No supertree equivalent possible for taxon or parent node
							if ($taxon =~ m/_1$/)	# Assign root taxon to any open clades
								{
								my $rootTaxon;
									($rootTaxon = $taxon) =~ s/_1$//;
								for (my $i = 1; $i <= $cladeNum; $i++)
									{
									my $clade = "GTnode" . $i;
									if ($open{$clade})
										{
										push @{ $geneNodeMember{$clade} }, $rootTaxon;
										$geneNodeMember{$rootTaxon}{$clade} = 1;
										}
									}
								}
							}
						else	# Taxon is a recognized singleton; assign to any open clades
							{
							$gene2ST{$taxon} = $taxon;
							for (my $i = 1; $i <= $cladeNum; $i++)
								{
								my $clade = "GTnode" . $i;
								if ($open{$clade})
									{
									push @{ $geneNodeMember{$clade} }, $taxon;
									$geneNodeMember{$taxon}{$clade} = 1;
									}
								}
							}
					$position += $offset-1;
					}
				$position++;
				}

			if ($debug)
				{
				foreach my $node (@GTnodeList)
					{
					print "\t\t\tClade ID: $node\n";
					if ($node =~ /GTnode/)
						{
						print "\t\t\t\tDescription: $CladeDescrip{$node}\n";
						print "\t\t\t\tNode descendants:";
						print " $_" foreach (@ { $geneNodeMember{$node}});
						print "\n";
						}
					print "\t\t\t\tBranch length: $CladeSupport{$node}\n";
					print "\t\t\t\tCorresponding supertree node: $gene2ST{$node}\n";
					print "\t\t\t\tParent: $ParentClade{$node}\n";
					if ($node =~ /GTnode/)
						{
						print "\t\t\t\tDaughters:";
						print " $_" foreach (@ { $daughterClade{$node}});
						print "\n";
						}
					}
				}

			# Determine rates of evolution for all branches
				my (%branchRate, %cladeRate);
				print "\t\tDetermining rates of evolution for all branches ...\n";
					print "\t\t\tClade statistics:\n" if ($verbose);
					
				foreach my $node (reverse @GTnodeList)	# Do in reverse order to enable determination of (nested) clade rates
					{
					$branchRate{$node} = $cladeRate{$node} = "undefined";
					
					if ($gene2ST{$node} eq "none")	# No corresponding ST clade due to multiple specimens; deal with separately
						{
						next unless ($node =~ /^GTnode/);	# Do not deal with terminal specimens directly

						my $realLength;
						foreach my $daughter (@ { $daughterClade{$node}})	# Calculate average branch length for specimens and add to node's branch length
							{
							$realLength += $CladeSupport{$daughter};
							}
						$realLength /= scalar(@ { $daughterClade{$node}});
						$CladeSupport{$node} += $realLength;
						($gene2ST{$node} = $daughterClade{$node}[1]) =~ s/_\d+$//;	# Redefine ST equivalent as that of root taxon

						$branchRate{$node} = $branchRate{$gene2ST{$node}} = $branchMatrix{$gene2ST{$node}}{$treeName{$treeID}} = $CladeSupport{$node} / ($nodeAge{$gene2ST{$ParentClade{$node}}} - $nodeAge{$gene2ST{$node}});
							$numGTrate++;
							$varGTrate += ($branchRate{$node} * $branchRate{$node});
						}
					else
						{
						# Get branch rate (except for root node and, as needed nodes descended from root node)
							unless ($node eq "GTnode1" or ($ParentClade{$node} eq "GTnode1" and $rootingStatus{$treeID} eq "[&U]") or $CladeSupport{$node} <= $lengthThreshold)
								{
								$branchRate{$node} = $branchRate{$gene2ST{$node}} = $branchMatrix{$gene2ST{$node}}{$treeName{$treeID}} = $CladeSupport{$node} / ($nodeAge{$gene2ST{$ParentClade{$node}}} - $nodeAge{$gene2ST{$node}});
									$numGTrate++;
									$varGTrate += ($branchRate{$node} * $branchRate{$node});
								}
								
						# Get clade rate (except for terminal taxa)
							if ($node =~ /^GTnode/)
								{
								$cladeRate{$node} = 0;
								my $numDaughters = 0;
								foreach my $daughter (@ { $daughterClade{$node}})
									{
									my $daughterRate = 0;
									my $numDaughterRates = 0;
									
									# Calculate rate for daughter lineage as average of branch rate and, if applicable, clade rate
										if (defined $branchRate{$daughter} and $branchRate{$daughter} ne "undefined")	# Get branch rate
											{
											$numDaughterRates++;
											$daughterRate += $branchRate{$daughter};
											}
										if (defined $cladeRate{$daughter} and $cladeRate{$daughter} ne "undefined")	# Get clade rate (if applicable)
											{
											$numDaughterRates++;
											$daughterRate += $cladeRate{$daughter};
											}
										if ($numDaughterRates > 0)	# Calculate average rate for daughter lineage and weight accordingly
											{
											$numDaughters++;
											$daughterRate /= $numDaughterRates;
											$daughterRate *= $GTcladeSize{$daughter} if ($sizeWeighting and defined $GTcladeSize{$daughter});
											$cladeRate{$node} += $daughterRate;
											}
									}
								
								if ($sizeWeighting)
									{
									$cladeRate{$node} /= $GTcladeSize{$node};
									}
								else
									{
									$cladeRate{$node} /= $numDaughters if ($numDaughters);
									}

								$cladeRate{$gene2ST{$node}} = $cladeMatrix{$gene2ST{$node}}{$treeName{$treeID}} = $cladeRate{$node};
								$geneRate{$treeName{$treeID}} = $cladeRate{$node} if ($node eq "GTnode1");	# Set rate for gene as clade rate for root node of gene tree
								}
						}

					# Print out node statistics
						if ($verbose)
							{
							print "\t\t\t\tClade ID: $node\n";
								print "\t\t\t\t\tDescription: $CladeDescrip{$node}\n" if ($node =~ /^GTnode/ and $debug);
							print "\t\t\t\t\tCorresponding supertree node: $gene2ST{$node}\n";
							print "\t\t\t\t\t\tAge of corresponding supertree node: ";
								if ($gene2ST{$node} eq "none")
									{
									print "n/a\n";
									}
								else
									{
									print "$nodeAge{$gene2ST{$node}}\n";
									}
							print "\t\t\t\t\tParent clade ID: ";
								if ($node eq "GTnode1")
									{
									print "none (basal clade)\n";
									}
								else
									{
									print "$ParentClade{$node} (= $gene2ST{$ParentClade{$node}})\n";
									print "$CladeDescrip{$ParentClade{$node}}\n" if ($debug);
									print "\t\t\t\t\t\tAge of corresponding supertree node: $nodeAge{$gene2ST{$ParentClade{$node}}}\n";
									}
							print "\t\t\t\t\tDaughter clade IDs: ";
								if (defined @ { $daughterClade{$node}})
									{
									foreach my $daughter (@ { $daughterClade{$node}})
										{
										print " $daughter";
										}
									}
								else
									{
									print "\tnone";
									}
								print "\n";
							print "\t\t\t\t\tBranch length: ";
								if (defined $CladeSupport{$node} and $node ne "GTnode1")
									{
									print "$CladeSupport{$node}\n";
									}
								else
									{
									print "none\n";
									}
							print "\t\t\t\t\t\tBranch specific rate of evolution: ";
								if ($branchRate{$node} eq "undefined")
									{
									print "$branchRate{$node}\n";
									}
								else
									{
									printf " %.3g\n", $branchRate{$node};
									}
							if ($node =~ /^GTnode/ and $gene2ST{$node} ne "none")
								{
								print "\t\t\t\t\tClade height: $cladeHeight{$node}\n";
								print "\t\t\t\t\t\tClade-specific (nested) rate of evolution: ";
								if ($cladeRate{$node} ne "undefined")
									{
									printf " %.3g\n", $cladeRate{$node};
									}
								else
									{
									print "undefined\n";
									}
								}
							}
					}

			# Derive mean rate of evolution for whole tree and each individual node
				if ($numGTrate > 1)
					{
					$varGTrate = ($varGTrate - ($geneRate{$treeName{$treeID}}**2 / $numGTrate)) / ($numGTrate - 1);
					$seGTrate = sqrt($varGTrate / $numGTrate); 
					}
				else
					{
					$varGTrate = $seGTrate = "n/a";
					}
				print TREE "$treeName{$treeID}\t$geneRate{$treeName{$treeID}}\t$seGTrate\t$numGTrate";
					print TREE "\n" unless ($windowAnalysis);
				printf "\n\t\tAverage nested rate of evolution across $treeName{$treeID}: %.3g (+/- %.3g; n = %s)\n", $geneRate{$treeName{$treeID}}, $seGTrate, $numGTrate if ($verbose);

			# Compare partition-specific rates to across partition rate
				if ($windowAnalysis)
					{
					my ($windowPairedN, $windowPlusPaired, $windowNegPaired, $windowZeroPaired) = (0, 0, 0, 0);
					my ($windowAvgPairedDiff, $windowVarPairedDiff, $windowSEPairedDiff);

					# Store information for each clade (for reference tree) or perform paired comparisons
						foreach my $iClade (@GTnodeList)
							{
							if ($treeName{$treeID} =~ /_all$/)
								{
								$windowMatrix{$treeName{$treeID}}{$gene2ST{$iClade}} = $branchRate{$iClade};	# Store across-partition value to allow future comparisons
								}
							else	# Compare to across partition tree
								{
								my $rootName;
									($rootName = $treeName{$treeID}) =~ s/_[a-z]+\d+$/_all/;
								next if (not defined $windowMatrix{$rootName}{$gene2ST{$iClade}} or $branchRate{$iClade} eq "undefined");	# Former conditional should also catch instances where tree based on all data does not preceed that based on a specific partition
								
								my $testValue = $branchRate{$iClade};
								my $refValue = $windowMatrix{$rootName}{$gene2ST{$iClade}};
									($testValue, $refValue) = lnTransform($testValue, $refValue) if ($logTransform);
								
								$windowPairedN++;
								my $windowPairedDiff = $testValue - $refValue;
									$windowAvgPairedDiff += $windowPairedDiff;
									$windowVarPairedDiff += $windowPairedDiff * $windowPairedDiff;
									
									$windowPlusPaired++ if ($windowPairedDiff > 0);
									$windowNegPaired++ if ($windowPairedDiff < 0);
								}
							}
							
					# Calculate statistical results for partition-specific analyses
						if ($treeName{$treeID} =~ /_all$/)
							{
							print TREE "\n";
							}
						else
							{
							my ($pairedStudentP, $pairedSignP);
							$windowAvgPairedDiff /= $windowPairedN if ($windowPairedN > 0);
		
							# Determine average relative rate et al.
								$windowVarPairedDiff = ($windowVarPairedDiff - ($windowAvgPairedDiff * $windowAvgPairedDiff) / $windowPairedN) / ($windowPairedN - 1);
									if ($windowVarPairedDiff <= 0)
										{
										$windowVarPairedDiff = 0;
										$windowSEPairedDiff = 0.001;
										}
									else
										{
										$windowSEPairedDiff = sqrt($windowVarPairedDiff / $windowPairedN);
										}
								
							# Determine if relative rate is significantly different from zero using two-sample t-test
								my $t = $windowAvgPairedDiff / $windowSEPairedDiff;
								my $df = $windowPairedN - 1;
					
								$pairedStudentP = _subtprob($df, $t);
									$pairedStudentP = 1 - $pairedStudentP if ($t < 0);
									$pairedStudentP *= 2;	# For two-tailed testing
					
								printf TREE "\t%.3g\t%.3g\t%.3g\t%.3g", $windowAvgPairedDiff, $windowSEPairedDiff, $t, $pairedStudentP;
					
							# Determine if relative rate is significantly different from zero using Fisher sign test
								my $signN = $windowPairedN - $windowZeroPaired;
								my $nTest = $windowPlusPaired;
								
								$pairedSignP = 2 * signTest($signN, $nTest);
								
								printf TREE "\t$windowPlusPaired\t$windowNegPaired\t%.3g\t%.3g", ($windowPlusPaired - $windowNegPaired) / ($windowPairedN - $windowZeroPaired), $pairedSignP;
								print TREE "\n";
							}
					}
					
			# Print results for each individual gene if desired
				if ($genePrint)
					{
					my $geneOutfile = $treeName{$treeID} . "_moleRat.txt";
					print "\t\tPrinting rates to file $geneOutfile ...\n";
					open (GENE, ">$geneOutfile") or die "Cannot write relative rates of evolution for $treeName{$treeID} to file $geneOutfile\n";
						# Print header row
							print GENE "Node\tNode age";
							print GENE "\tBranch rate\tTree average\tSE\tt\tt p-value";	# t-test branch header
								print GENE "\tRate(sig)\tRate(ns)\tRate(n/a)" if ($graphPrint);
							print GENE "\tClade rate\tTree average\tSE\tt\tt p-value";	# t-test clade header
								print GENE "\tAverage rate(sig)\tAverage rate(ns)\tAverage rate(n/a)" if ($graphPrint);
							print GENE "\n";
						
					foreach my $clade (@STnodeList) 
						{
						# Skip clades without a branch- or clade-specific rate
							next if (not defined $branchRate{$clade} and not defined $cladeRate{$clade});
							next if (defined $branchRate{$clade} and $branchRate{$clade} eq "undefined" and defined $cladeRate{$clade} and $cladeRate{$clade} eq "undefined");

						print GENE "$clade\t$nodeAge{$clade}";
		
						# Print branch-specific rates
							if (not defined $branchRate{$clade})
								{
								print GENE "\t\t\t\t\t";
								print GENE "\t\t\t" if ($graphPrint);
								}
							else
								{
								printf GENE "\t%.3g\t%.3g\t%.3g", $branchRate{$clade}, $geneRate{$treeName{$treeID}}, $seGTrate;
								
								# Perform one-sample t-test
									my $t = ($branchRate{$clade} - $geneRate{$treeName{$treeID}}) / $seGTrate;
									my $df = $numGTrate;
									my $tProb = _subtprob($df, $t);
										$tProb = 1 - $tProb if ($t < 0);
										$tProb *= 2;	# For two-tailed testing
									
								printf GENE "\t%.3g\t%.3g", $t, $tProb;
									if ($graphPrint)
										{
										if ($numGTrate == 1)	# Insufficient sample size
											{
											printf GENE "\t\t\t%.3g", $branchRate{$clade};
											}
										elsif ($tProb < 0.05)	# Significant result
											{
											printf GENE "\t%.3g\t\t", $branchRate{$clade};
											}
										else	# ($tProb >= 0.05) Non-significant result
											{
											printf GENE "\t\t%.3g\t", $branchRate{$clade};
											}
										}
								}
					
						# Print clade-specific results
							if (defined $cladeRate{$clade} and $cladeRate{$clade} ne "undefined")	# Skip terminal branches and nodes without clade rates
								{
								printf GENE "\t%.3g\t%.3g\t%.3g", $cladeRate{$clade}, $geneRate{$treeName{$treeID}}, $seGTrate;
								
								# Perform one-sample t-test
									my $t = ($cladeRate{$clade} - $geneRate{$treeName{$treeID}}) / $seGTrate;
									my $df = $numGTrate;
									my $tProb = _subtprob($df, $t);
										$tProb = 1 - $tProb if ($t < 0);
										$tProb *= 2;	# For two-tailed testing
									
								printf GENE "\t%.3g\t%.3g", $t, $tProb;
									if ($graphPrint)
										{
										if ($numGTrate == 1)	# Insufficient sample size
											{
											printf GENE "\t\t\t%.3g", $cladeRate{$clade};
											}
										elsif ($tProb < 0.05)	# Significant result
											{
											printf GENE "\t%.3g\t\t", $cladeRate{$clade};
											}
										else	# ($tProb >= 0.05) Non-significant result
											{
											printf GENE "\t\t%.3g\t", $cladeRate{$clade};
											}
										}
								}
							print GENE "\n";
						}
					close GENE;
					}
			printf "\t\tTime taken: %s seconds\n", time - $geneZero;
			}
	close TREE;
	printf "\n\tTotal time taken: %s seconds\n", time - $allGeneZero;

exit(0) unless ($branchTests);
exit(0) if ($windowAnalysis);

# Determine branches or clades with rates that are significantly different from average for tree ("absolute") or that have shifted significantly from a nearby reference ancestor ("shifts") using paired statistical tests (Student's t-test or Fisher's sign test)
	my $statsZero = time;
	print "\nDetermining branches and clades with significantly altered rates of evolution ...\n";
	
	# Print header rows
		# Absolute gene rates for branches and clades
			open (MATRIX1, ">branchMatrix_moleRat.txt") or die "Cannot write relative rates of evolution for branches to file branchMatrix_moleRat.txt\n";
			open (MATRIX2, ">cladeMatrix_moleRat.txt") or die "Cannot write relative rates of evolution for clades to file cladeMatrix_moleRat.txt\n";
				print MATRIX1 "Node";
					foreach my $gene (@GTnames)
						{
						print MATRIX1 "\t$gene";
						print MATRIX2 "\t$gene";
						}
					print MATRIX1 "\n";
					print MATRIX2 "\n";
				print MATRIX1 "Gene average";
				print MATRIX2 "Gene average";
					foreach my $gene (@GTnames)
						{
						printf MATRIX1 "\t%.3g", $geneRate{$gene};
						printf MATRIX2 "\t%.3g", $geneRate{$gene};
						}
					print MATRIX1 "\n";
					print MATRIX2 "\n";
	
		# Branch-specific differences and shifts
			open (BRANCH, ">branchRates_moleRat.txt") or die "Cannot write relative rates of evolution for branches to file branchRates_moleRat.txt\n";
				print BRANCH "Node\tNode age\tn";
				print BRANCH "\tAverage paired difference\tSE\tt\tt p-value";	# t-test header
					print BRANCH "\tAverage paired difference(sig)\tAverage paired difference(ns)\tAverage paired difference(n/a)" if ($graphPrint);
				print BRANCH "\tplusCount\tnegCount\tProportion\tsign p-value";	# Sign test header
					print BRANCH "\tProportion(sig)\tProportion(ns)\tProportion(n/a)" if ($graphPrint);
				print BRANCH "\n";
			close BRANCH;
		
			open (SHIFT1, ">branchShifts_moleRat.txt") or die "Cannot write shifts in relative rates of evolution for branches to file branchShifts_moleRat.txt\n";
				print SHIFT1 "Node\tNode age\tPaired n\tReference ancestor\tDepth to ancestor";
				print SHIFT1 "\tAverage paired difference\tSE\tPaired t\tPaired t p-value";	# t-test header
					print SHIFT1 "\tAverage paired difference(sig)\tAverage paired difference(ns)\tAverage paired difference(n/a)" if ($graphPrint);
				print SHIFT1 "\tplusCount\tnegCount\tProportion\tPaired sign p-value";	# Sign test header
					print SHIFT1 "\tProportion(sig)\tProportion(ns)\tProportion(n/a)" if ($graphPrint);
				print SHIFT1 "\n";
			close SHIFT1;

		# Clade-specific differences and shifts
			open (CLADE, ">cladeRates_moleRat.txt") or die "Cannot write relative rates of evolution for clades to file cladeRates_moleRat.txt\n";
				print CLADE "Node\tNode age\tSize\tn";
				print CLADE "\tAverage paired difference\tSE\tt\tt p-value";	# t-test header
					print CLADE "\tAverage paired difference(sig)\tAverage paired difference(ns)\tAverage paired difference(n/a)" if ($graphPrint);
				print CLADE "\tplusCount\tnegCount\tProportion\tsign p-value";	# Sign test header;
					print CLADE "\tProportion(sig)\tProportion(ns)\tProportion(n/a)" if ($graphPrint);
				print CLADE "\n";
			close CLADE;
	
			open (SHIFT2, ">cladeShifts_moleRat.txt") or die "Cannot write shifts in relative rates of evolution for clades to file cladeShifts_moleRat.txt\n";
				print SHIFT2 "Node\tNode age\tSize\tPaired n\tReference ancestor\tDepth to ancestor";
				print SHIFT2 "\tAverage paired difference\tSE\tPaired t\t Paired t p-value";	# t-test header
					print SHIFT2 "\tAverage paired difference(sig)\tAverage paired difference(ns)\tAverage paired difference(n/a)" if ($graphPrint);
				print SHIFT2 "\tplusCount\tnegCount\tProportion\tPaired sign p-value";	# Sign test header;
					print SHIFT2 "\tProportion(sig)\tProportion(ns)\tProportion(n/a)" if ($graphPrint);
				print SHIFT2 "\n";
			close SHIFT2;

		my (%branchCount, %cladeCount);
			$branchCount{"STnode0"} = $cladeCount{"STnode0"} = 0;
		foreach my $clade (@STnodeList)
			{
			print "\tClade ID: $clade\n" if ($verbose);
			$branchCount{$clade} = $cladeCount{$clade} = 0;
			
			# Print to matrix output files
				print MATRIX1 "$clade";
				print MATRIX2 "$clade" if ($clade =~ /STnode/);
					foreach my $gene (@GTnames)
						{
						print MATRIX1 "\t";
						print MATRIX2 "\t" if ($clade =~ /STnode/);
						if (defined $branchMatrix{$clade}{$gene})
							{
							printf MATRIX1 "%.3g", $branchMatrix{$clade}{$gene};
							$branchCount{$clade}++;
							}
						next unless ($clade =~ /STnode/);
						if (defined $cladeMatrix{$clade}{$gene})
							{
							printf MATRIX2 "%.3g", $cladeMatrix{$clade}{$gene};
							$cladeCount{$clade}++;
							}
						}
					print MATRIX1 "\n";
					print MATRIX2 "\n" if ($clade =~ /STnode/);
				if ($debug)
					{
					print "\t\t$branchCount{$clade} gene(s) providing branch-specific rates\n";
					print "\t\t$cladeCount{$clade} gene(s) providing clade-specific rates\n";
					}

			# Perform paired comparisons for branch- and clade-specific rates
				foreach my $outFile (@fileNames)	# Loop over analysis types (branch- or clade-specific rates, each each either "average" or "shifts")
					{
					if ($outFile =~ /clade/i)
						{
						next unless ($clade =~ /^STnode/);
						}
					print "\t\tAnalysis: $outFile\n" if ($verbose);

					# Clear variables
						my ($avgDiff, $varDiff) = (0, 0);
						my ($pairedN, $plusCount, $negCount) = (0, 0, 0);

					# Get reference ancestors for shift analyses
						my $refClade = "STnode0";	# Default to "undefined" reference
						my $refDepth = 0;
							my $testRefDepth = 0;
						my $testRef = $clade;
						my $maxMatch = 0;
						
						my $testCount;
							$testCount = $branchCount{$clade} if ($outFile =~ /^branch/i);
							$testCount = $cladeCount{$clade} if ($outFile =~ /^clade/i);

						if ($outFile =~ /shift/i)
							{
							unless ($testCount == 0 or $clade eq "STnode1")	# Branch has no data or already at root of tree
								{
								print "\t\t\tDetermining reference branch ...\n" if ($debug);
								until ($testRef eq "STnode0" or ($depthThreshold and $testRefDepth >= $depthThreshold))	# Find (closest) ancestral branch with most paired data
									{
									$testRef = $ParentClade{$testRef};
									$testRefDepth++;
									print "\t\t\t\tComparing against $testRef ($testRefDepth)\n" if ($debug);
									
									# Automatically skip references without data
										next if ($outFile =~ /^branch/i and $branchCount{$testRef} == 0);
										next if ($outFile =~ /^clade/i and $cladeCount{$testRef} == 0);
									
									# Get number of paired comparisons across genes
										my $matchCount= 0;
										foreach my $gene (@GTnames)
											{
											$matchCount++ if ($outFile =~ /^branch/i and defined $branchMatrix{$clade}{$gene} and defined $branchMatrix{$testRef}{$gene});
											$matchCount++ if ($outFile =~ /^clade/i and defined $cladeMatrix{$clade}{$gene} and defined $cladeMatrix{$testRef}{$gene});
											}
										print "\t\t\tNumber of paired matches: $matchCount\n" if ($debug);
									if ($matchCount > $maxMatch)
										{
										$refClade = $testRef;
										$maxMatch = $matchCount;
										$refDepth = $testRefDepth;
										print "\t\t\t\tNew reference branch\n" if ($debug);
										}
									}
								}
							print "\t\t\tReference branch: $refClade (depth = $refDepth; n = $maxMatch)\n" if ($debug);
							}
						else
							{
							$refClade = "n/a";
							}
					
					# Do paired comparisons
						unless ($outFile =~ /shift/i and $refClade eq "STnode0")	# Skip shift analyses when reference clade is undefined
							{
							foreach my $gene (@GTnames)
								{
								# Get respective values for each analysis type
									my $testValue;
										$testValue = $branchMatrix{$clade}{$gene} if ($outFile =~ /^branch/i and defined $branchMatrix{$clade}{$gene});
										$testValue = $cladeMatrix{$clade}{$gene} if ($outFile =~ /^clade/i and defined $cladeMatrix{$clade}{$gene});
										
									my $refValue;
										if ($outFile =~ /shift/i)
											{
											$refValue = $branchMatrix{$refClade}{$gene} if ($outFile =~ /^branch/i and defined $branchMatrix{$refClade}{$gene});
											$refValue = $cladeMatrix{$refClade}{$gene} if ($outFile =~ /^clade/i and defined $cladeMatrix{$refClade}{$gene});
											}
										else
											{
											$refValue = $geneRate{$gene};
											}

								# Do paired comparisons (really!)
									if (defined $testValue and defined $refValue and $refClade ne "STnode0")
										{
										($testValue, $refValue) = lnTransform($testValue, $refValue) if ($logTransform);
										$pairedN++;
										my $pairedDiff = $testValue - $refValue;
											$avgDiff += $pairedDiff;
											$varDiff += ($pairedDiff * $pairedDiff);
											
											$plusCount++ if ($pairedDiff > 0);
											$negCount++ if ($pairedDiff < 0);
										}
								}
							}
						$avgDiff /= $pairedN if ($pairedN > 0);
						printStats($outFile, $pairedN, $plusCount, $negCount, $avgDiff, $varDiff, $clade, $refClade, $refDepth);
					}
			}
 			
	close MATRIX1;
	close MATRIX2;
	printf "\tTime taken: %s seconds\n", time - $statsZero;

printf "\nTotal time taken: %s seconds\n", time - $rateZero;
	
exit(0);

### Subroutines used in script

sub treeRead	# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift;

	setLineBreak($treeFile);
	open (TREES, "<$treeFile") or die "Cannot open nexus-formatted treefile $treeFile\n";
		my (@allTrees, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount, $rooting);
		my (@taxonList, %taxonPresent);
		my (%TTable, $translation, $translateFlag,$maxTaxonID, $transText);
		my $commentFlag = 0;
		
		while (<TREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			$_ =~ s/\s+$//;		# Strip (any) trailing white space
			
			# Block any commented lines
				if ($_ =~ /^\[/)
					{
					$commentFlag += ($_ =~ tr/\[+//);
					}
				if ($_ =~ /\]\s?$/)
					{
					$_ =~ s/\s+$//;
					$commentFlag -= ($_ =~ tr/\]+//);
					next;
					}
				next if ($commentFlag > 0);
				
			# Process various nexus commands
				if ($_ =~ /^begin trees;/i)		# Clear variables for each tree block
					{
					undef %TTable;
					$translation = $translateFlag = $treeFlag = $treeCount = 0;
					$treeBlock++;
					}
				if ($_ =~ /^translate/i)	# Read translation table of taxon names (if any)
					{
					print "Reading translation table ...\n" if ($verbose);
					$translateFlag = $translation = 1;
					$transText = "";
					$maxTaxonID = 0;
					}
				if ($translateFlag)
					{
					unless ($_ =~ /^translate/i)
						{
						$transText .= $_;
						if ($_ =~ /;/)	# End of translate command; parse
							{
							$translateFlag = 0;
							$transText =~ s/;$//;	#Strip trailing semi-colon
							my @transEntries = split(/,/, $transText);
							foreach my $entry (@transEntries)
								{
								$entry =~ s/^\s+//;		#Strip (any) leading white space
								my @transEntry = split(/\s/,$entry);
									my $taxonID = shift(@transEntry);
									my $taxonName = join(' ', @transEntry);

								$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
								$taxonName =~ s/\s+/_/g;
									$taxonName = "'" . $taxonName . "'" if ($taxonName =~ /\W/);
									$taxonName =~ s/^\'+/\'/;
									$taxonName =~ s/\'+$/\'/;
								$TTable{$taxonID} = $taxonName;
								$taxonPresent{$taxonName}++;
								push @taxonList, $taxonName;
								}
							}
						}
					}
				if (($_ =~ /tree/i and $_ =~ /=/) or $_ =~ /^\(/)
					{
					unless ($_ =~ /^>/)		#Ignore (any) commented lines
						{
						$rawTreeDescrip = $treeDescrip = "";
						unless ($block == 0 and $treeBlock > 1)
							{
							$treeFlag = 1;
							}
						}
					}
	
				if ($treeFlag)
					{
					$rawTreeDescrip .= $_;
					if ($_ =~ /:/)
						{
						$SupportFlag = 1;
						}
					if ($_ =~ /;/)	# End of tree description; parse
						{
						$treeCount++;
						print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
						print "\tTranslating taxon names ...\n" if ($verbose and $translation);
						# Determine if trees are rooted or not (default for no information is not rooted)
							if ($rawTreeDescrip =~ /\[&R\]/)
								{
								$rooting = "[&R]";
								}
							elsif ($rawTreeDescrip =~ /\[&U\]/)
								{
								$rooting = "[&U]";
								$unrootedFlag = 1;
								}
							else
								{
								$rooting = $defaultRooting;
								$unrootedFlag = 1 if ($defaultRooting eq "[&U]");
								}
						# Remove any embedded comments
							if ($rawTreeDescrip =~ /\[/)
								{
								for (my $position = 0; $position < length($rawTreeDescrip); $position++)
									{
									if (substr($rawTreeDescrip, $position, 1) eq "[")
										{
										$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
										}
									else
										{
										$treeDescrip .= substr($rawTreeDescrip, $position, 1);
										}
									}
								}
							else
								{
								$treeDescrip = $rawTreeDescrip;
								}
						$treeFlag = 0;
						my $position = 0;
						until (substr($treeDescrip, $position, 1) eq "(")		# Avoid anything before tree description proper
							{
							$position++;
							}
						my $tree = substr($treeDescrip, $position);
							$tree =~ s/;\s*$//;
							$tree =~ s/\s+//g;
							
						my $name;
						if ($treeDescrip =~ /=/)
							{
							($name) = (split /=/, $treeDescrip, 0);		# Get name of tree, if any
							$name =~ s/\s*\w*tree\s+//i;
							$name =~ s/\s*\*\s*//i;
							$name =~ s/\s+$//g;
							$name =~ s/-/_/g;
#							$treeNames{$treeCount} = $name;
							}
						my $openCount = ($tree =~ tr/(//);
						my $closedCount = ($tree =~ tr/)//);
						if ($openCount == $closedCount)		# Tree is balanced; process
							{
							$position = length($tree) - 1;
							while ($position >= 0)
								{
								my $IDlength = 1;
								if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
									{
									$position--;
									}
								else	# Read in taxonID
									{
									until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
										{
										$position--;
										$IDlength++;
										}
									unless (substr($tree, $position, 1) eq ":" or substr($tree, $position, 1) eq ")")		# Avoid reading branch lengths or node labels as taxon names
										{
										my $readName = substr($tree, $position+1, $IDlength-1);
										if ($translation and not defined $taxonPresent{$readName})	# Translate taxon names; latter condition is a safety for BioPerl convert_trees.pl module which adds, but doesn't use translation block
											{
											if ($TTable{$readName})
												{
												substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		# Translate name
												unless ($taxonPresent{$TTable{$readName}})
													{
													push @taxonList, $TTable{$readName};
													$taxonPresent{$TTable{$readName}} = 1;
													}
												}
											else
												{
												print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
												substr($tree, $position+1, $IDlength-1) = "xxx";		# Translate name
												unless ($taxonPresent{"xxx"})
													{
													push @taxonList, "xxx";
													$taxonPresent{"xxx"} = 1;
													}
												}
											}
										else	# Enter name as read from tree, unless already present
											{
											$taxonPresent{$readName}++;
											push @taxonList, $readName if ($taxonPresent{$readName} == 1);
											}
										}
									}
								}
								
							$treeNum++;
								my $treeID = "treeID_$treeNum";
								$name = "readTree_$treeNum" if (not defined $name);
							
							push @allTrees, $treeID;
								$treeDescription{$treeID} = $tree;
								$rootingStatus{$treeID} = $rooting;
								$treeName{$treeID} = $name;
							print "\tTree description:\n\t\t$tree\n" if ($verbose);
							}
						else	#Tree is unbalanced; reject
							{
							print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
							}
						}
					}
			}
	close TREES;
	
	# Process taxon list
		@allTaxa = sort(duplStrip(@taxonList));
	
	return @allTrees;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub _subtprob	# Calculates upper probability from Student's t-distribution; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($n, $x) = @_;

	my ($a,$b);
	my $w = atan2($x / sqrt($n), 1);
	my $z = cos($w) ** 2;
	my $y = 1;

	for (my $i = $n-2; $i >= 2; $i -= 2)
		{
		$y = 1 + ($i-1) / $i * $z * $y;
		} 

	if ($n % 2 == 0)
		{
		$a = sin($w)/2;
		$b = .5;
		}
	else
		{
		$a = ($n == 1) ? 0 : sin($w)*cos($w)/PI;
		$b= .5 + $w/PI;
		}
	return max(0, 1 - $b - $a * $y);
	}
	
sub max	# Finds maximum value in an array; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my $max = shift;
	my $next;
	while (@_)
		{
		$next = shift;
		$max = $next if ($next > $max);
		}	
	return $max;
	}

sub signTest	# Calculates one-tailed p-value for sign test
	{
	my $n = shift;
	my $k = shift;
		die "ERROR: n ($n) > k ($k)\n" if ($k > $n);
		my ($lowerK, $upperK) = ($k, $k);	# Upper k used for exact and normal calculations; lower k for others
			if ($k > $n / 2)
				{
				$lowerK = $n - $k;
				}
			else
				{
				$upperK = $n - $k;
				}

	my $binomialP = 0;
	
	if ($n <= 60)	# Use stored values of n! directly
		{
		if (not defined $factorial{0})	# Store values if not already present
			{
			$factorial{0} = 1;
			$factorial{$_} = $factorial{$_ - 1} * $_ for (1..60);	# Quickly store factorials up to N for subsequent sign tests
			}

		for (my $coeffic = $upperK; $coeffic <= $n; $coeffic++)
			{
			$binomialP += $factorial{$n} / ($factorial{$coeffic} * $factorial{$n - $coeffic});
			}
		$binomialP /= 2**$n;
		printf "The cumulative probabilities for the binomial $n choose $k (calculated directly) are:\n\t%.3g (one-tailed)\n\t%.3g (two-tailed)\n\n", $binomialP, $binomialP * 2 if ($debug);
		}
	elsif ($n <= 1000)	# Use gamma function to approximate n!
		{
		# For probability of binomial distribution of n and k calculated iteratively (from a hint in Numerical Recipes in C)
			my $rootBinomialCo = int(0.5 + exp(lnGammaApprox($n + 1) - lnGammaApprox(0 + 1) - lnGammaApprox($n - 0 + 1)));	# From Numerical Recipes in C
			my $rootBinomialP = $rootBinomialCo;
				$rootBinomialP /= 2 for (1..$n);
				$binomialP += $rootBinomialP;
			print "Root: $rootBinomialCo\t$rootBinomialP\n" if ($debug);
			
			for (my $i = 0; $i < $lowerK; $i++)
				{
				my $iBinomialCo = $rootBinomialCo * ($n - $i) / ($i + 1);
				my $iBinomialP = $iBinomialCo;
					$iBinomialP /= 2 for (1..$n);
				print "\tIteration $i: $iBinomialCo\t$iBinomialP\n" if ($debug);
				
				$rootBinomialCo = $iBinomialCo;
				$binomialP += $iBinomialP;
				}
			printf "The cumulative probabilities for the binomial $n choose $k (calculated using recurrence relations) are:\n\t%.3g (one-tailed)\n\t%.3g (two-tailed)\n\n", $binomialP, $binomialP * 2 if ($debug);
		
		# Calculated brute force
			$binomialP = 0;
			for (0..$lowerK)
				{
				my $k = $_;
				printf "The binomial coefficient for $n choose $k is %s\n", int(0.5 + exp(lnGammaApprox($n + 1) - lnGammaApprox($k + 1) - lnGammaApprox($n - $k + 1))) if ($debug);	# From Numerical Recipes in C
				printf "The probability for the binomial $n choose $k is %.3g\n", int(0.5 + exp(lnGammaApprox($n + 1) - lnGammaApprox($k + 1) - lnGammaApprox($n - $k + 1))) / 2**$n if ($debug);
				$binomialP += int(0.5 + exp(lnGammaApprox($n + 1) - lnGammaApprox($k + 1) - lnGammaApprox($n - $k + 1))) / 2**$n;
				}
			printf "The cumulative probabilities for the binomial $n choose $k (calculated brute force) are:\n\t%.3g (one-tailed)\n\t%.3g (two-tailed)\n\n", $binomialP, $binomialP * 2 if ($debug);
		}
	else	# Use normal approximation
		{
		my $p = $upperK;
		my $q = $n - $p;
		
		my $Z = (abs($p - ($n * 0.5)) - 0.5) / (sqrt($n * ($p / $n) * ($q / $n)));	# Slightly modified from Zar (uses $p in denominator in place of $n)
		
		$binomialP = _subuprob($Z);
		printf "The cumulative probabilities for the binomial $n choose $k (calculated using normal approximation) are:\n\t%.3g (one-tailed)\n\t%.3g (two-tailed)\n\n", $binomialP, $binomialP * 2 if ($debug);
		}

	$binomialP = 0.5 if ($binomialP > 0.5);	
	return $binomialP;
	}

sub lnGammaApprox	# Approximates ln gamma function using Lanczos approximation (using algorithm from Numerical Recipes in C)
	{
	my $z = shift;
	my %p = ('0' => '1.000000000190015',
			 '1' => '76.18009172947146',
			 '2' => '-86.50532032941677',
			 '3' => '24.01409824083091',
			 '4' => '-1.231739572450155',
			 '5' => '1.208650973866179e-3',
			 '6' => '-5.395239384953e-6');
	my ($x, $y);
		$x = $y = $z;

	my $tmp = $x + 5.5;
		$tmp -= ($x + 0.5) * log($tmp);
	my $ser = $p{0};
	for (1..6)
		{
		$ser += $p{$_}/++$y;
		}
	my $lnGamma = (-1 * $tmp) + log(2.5066282746310005 * $ser/$x);
	return $lnGamma;
	}

sub _subuprob	# Calculates upper probability from normal distribution; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($x) = @_;
	my $p = 0;
	my $absx = abs($x);

	if ($absx < 1.9)
		{
		$p = (1 +
			$absx * (.049867347
			  + $absx * (.0211410061
			  	+ $absx * (.0032776263
				  + $absx * (.0000380036
					+ $absx * (.0000488906
					  + $absx * .000005383)))))) ** -16/2;
		}
	elsif ($absx <= 100)
		{
		for (my $i = 18; $i >= 1; $i--)
			{
			$p = $i / ($absx + $p);
			}
		$p = exp(-.5 * $absx * $absx) / sqrt(2 * PI) / ($absx + $p);
		}
	else
		{
		$p = 0;
		}

	$p = 1 - $p if ($x < 0);
	return $p;
	}
	
sub printStats	# Calculate stats for Student's t and sign test and output to file
	{
	# Get variables
		my $statsFile = shift;
		my $printN = shift;
			my $printPlus = shift;
			my $printNeg = shift;
			my $printZero = $printN - ($printPlus + $printNeg);
		my $printAvg = shift;
			my $printVar = shift;
			my $printSE;
		my $printClade = shift;
			my $printAge = $nodeAge{$printClade};
			my $printSize = sprintf("%s", $STcladeSize{$printClade}) if ($printClade =~ /STnode/);
			my $printRef = shift;
				$printRef = "n/a" if (not defined $printRef);
			my $printDepth = shift;			
				$printDepth = "n/a" if (not defined $printDepth);
		my ($printStudentP, $printSignP);

	open (STATS, ">>$statsFile") or die "Cannot write results to file $statsFile\n";
		if ($printN == 0 or $printRef eq "STnode0")	# or $printRef == 1 or ($printPlus + $printNeg == 0)
			{
			$printAvg = $printSE = $printStudentP = $printSignP = "n/a";
			$printRef = $printDepth = $printN = "n/a" if ($printRef eq "STnode1");
			
			if ($verbose)
				{
				print STATS "$printClade\t$printAge";
					print STATS "\t$printSize" if ($statsFile =~ /clade/);
					print STATS "\t$printN";
					print STATS "\t$printRef\t$printDepth" if ($statsFile =~ /Shifts/);
					print STATS "\n";
				print "\t\t\tNumber of estimates: $printN\n";
				printf "\t\t\tAverage paired difference: %s (p = %s)\n", $printAvg, $printStudentP;
				printf "\t\t\t%s positive versus %s negative rates (p = %s)\n", $printPlus, $printNeg, $printSignP;
				}
			}
		elsif ($printN == 1)
			{
			$printSE = $printStudentP = $printSignP = "n/a";
			my $plusProportion = "";
				$plusProportion = sprintf("%.3g", ($printPlus - $printNeg) / ($printN - $printZero)) if ($printN - $printZero > 0);
	
			print STATS "$printClade\t$printAge";
				print STATS "\t$printSize" if ($statsFile =~ /clade/);
				print STATS "\t$printN";
				print STATS "\t$printRef\t$printDepth" if ($statsFile =~ /Shifts/);
			printf STATS "\t%.3g\t\t\t", $printAvg;
				printf STATS "\t\t\t%.3g", $printAvg if ($graphPrint);
			printf STATS "\t$printPlus\t$printNeg\t%s\t", $plusProportion;
				printf STATS "\t\t\t%s", $plusProportion if ($graphPrint);
			print STATS "\n";
	
			if ($verbose)
				{
				print "\t\t\tNumber of estimates: $printN\n";
				printf "\t\t\tAverage paired difference: %.3g (p = %s)\n", $printAvg, $printStudentP;
				printf "\t\t\t%s positive versus %s negative rates (p = %s)\n", $printPlus, $printNeg, $printSignP;
				}
			}
		else
			{
			print STATS "$printClade\t$printAge";
				print STATS "\t$printSize" if ($statsFile =~ /clade/);
				print STATS "\t$printN";
				print STATS "\t$printRef\t$printDepth" if ($statsFile =~ /Shifts/);
			# Determine average relative rate et al.
				$printVar = ($printVar - ($printAvg * $printAvg) * $printN) / ($printN - 1);
					if ($printVar <= 0)
						{
						$printVar = 0;
						$printSE = 0.001;
						}
					else
						{
						$printSE = sqrt($printVar / $printN);
						}
				
			# Determine if relative rate is significantly different from zero using t-test
				my $t = $printAvg / $printSE;
				my $df = $printN - 1;
	
				$printStudentP = _subtprob($df, $t);
					$printStudentP = 1 - $printStudentP if ($t < 0);
					$printStudentP *= 2;	# For two-tailed testing
	
				printf STATS "\t%.3g\t%.3g\t%.3g\t%.3g", $printAvg, $printSE, $t, $printStudentP;
					if ($graphPrint)
						{
						printf STATS "\t%.3g\t\t", $printAvg if ($printStudentP < 0.05);
						printf STATS "\t\t%.3g\t", $printAvg if ($printStudentP > 0.05);
						}
	
			# Determine if relative rate is significantly different from zero using Fisher sign test
				my $signN = $printN - $printZero;
				my $nTest = $printPlus;
				my $plusProportion = "";
				
				if ($signN == 0)
					{
					$printSignP = "n/a";
					}
				else
					{
					$printSignP = sprintf("%.3g", 2 * signTest($signN, $nTest));
					$plusProportion = sprintf("%.3g", ($printPlus - $printNeg) / $signN);
					}

				printf STATS "\t$printPlus\t$printNeg\t%s\t%s", $plusProportion, $printSignP;
					if ($graphPrint)
						{
						if ($signN <= 5)	# Sample size too small to give significant result
							{
							printf STATS "\t\t\t%s", $plusProportion;
							}
						elsif ($printSignP < 0.05)	# Significant result
							{
							printf STATS "\t%s\t\t", $plusProportion;
							}
						else	# ($printSignP >= 0.05) Non-significant results
							{
							printf STATS "\t\t%s\t", $plusProportion;
							}
						}
				print STATS "\n";
	
			if ($verbose)
				{
				print "\t\t\tNumber of estimates: $printN\n";
				printf "\t\t\tAverage paired difference: %.3g +/- %.3g (t = %.3g; p = %.3g)\n", $printAvg, $printSE, $t, $printStudentP;
				printf "\t\t\t%s positive versus %s negative rates (p = %.3g)\n", $printPlus, $printNeg, $printSignP;
				}
			}
	close STATS;
	return;
	}

sub partitionMetric	# NOTE: Uses version v1.1
	{
	my @treeID = @_;
	my $finalTree = scalar(@treeID - 2);
		$finalTree = 0 if ($compType eq "ref");
	my ($dS, $norm_dS);
	
	if ($debug)
		{
		print "Calculating ";
		print "weighted " if ($weighted_dS);
		print "partition metrics ...\n\n";
		}
	
	for (my $treeCount1 = 0; $treeCount1 <= $finalTree; $treeCount1++)
		{
		my $treeID1 = $treeID[$treeCount1];
		my $tree1 = $treeDescription{$treeID1};
		$rootingStatus{$treeID1} = $defaultRooting if (not defined $rootingStatus{$treeID1});
		print "Comparing versus tree $treeCount1 ($treeID1)\n\t$tree1\n" if ($debug);
		
		for (my $treeCount2 = $treeCount1 + 1; $treeCount2 < scalar(@treeID); $treeCount2++)
			{
			my $treeID2 = $treeID[$treeCount2];
			my $tree2 = $treeDescription{$treeID2};

			$rootingStatus{$treeID2} = $defaultRooting if (not defined $rootingStatus{$treeID2});
			print "\tComparing versus tree $treeCount2 ($treeID2)\n\t\t$tree2\n" if ($debug);
			
			$rootingStatus = "unrooted";
				$rootingStatus = "rooted" if ($rootingStatus{$treeID1} eq "[&R]" and $rootingStatus{$treeID2} eq "[&R]");
				print "\n\tComparison type: $rootingStatus\n\n" if ($debug);
			
			# Prune trees to identical taxon sets
				pruneTree($treeID1, $treeID2);
					$tree1 = $prunedTree{$treeID1};
					$tree2 = $prunedTree{$treeID2};
					
					# Get pruned tree size for normalized dS and check whether to continue
						my $taxNum;
							if ($tree1 eq "()")
								{
								$taxNum = 0;
								}
							else
								{
								$taxNum = ($tree1 =~ tr/\,//) + 1;
								}
							$pruneSize{$treeID1}{$treeID2} = $taxNum;

			# Calculate dS
				if (($taxNum < 4 and $rootingStatus eq "unrooted") or ($taxNum < 3 and $rootingStatus eq "rooted"))	# Not enough shared taxa for a meaningful comparison
					{
					$dS = $norm_dS = $dSlist{$treeID1}{$treeID2} = $norm_dSlist{$treeID1}{$treeID2} = "n/a";
					}
				else
					{
					undef %cladeWeight;
					my @dotPlots;
						push @dotPlots, dotPlotter($rootingStatus, $tree1);	# Not as efficient rederiving dot plots each time for first tree (could be done once in outer loop), but necessary for storing clade weights
						push @dotPlots, dotPlotter($rootingStatus, $tree2);
		
					my %plotCount;
					$dS = 0;
					
					print "\nDot plots from trees ...\n" if ($debug);
					foreach my $plot (@dotPlots)
						{
						$plotCount{$plot}++;
						if ($plotCount{$plot} == 1)	# First incidence of clade; increment dS
							{
							$dS += $cladeWeight{$plot}; 
							}
						$dS -= $cladeWeight{$plot} if ($plotCount{$plot} == 2);	# Clade appears in both trees (i.e., not unique); de-increment dS
							$dS = 0 if ($dS < 10**-5);	# Correction because floating point calculations can occasionally give a slightly negative number (i.e., on the order of -1e-16)
						print "$plot ($plotCount{$plot}; dS = $dS)\n" if ($debug);
						}
					print "\n\tTree $treeCount1 vs. tree $treeCount2: dS = $dS\n\n" if ($debug);
					$dSlist{$treeID1}{$treeID2} = $dS;

					# Calculate normalized dS and store
						if ($rootingStatus eq "rooted")
							{
							$norm_dS = $norm_dSlist{$treeID1}{$treeID2} = sprintf ("%.3f", $dS / ((2 * $taxNum) - 4));
							}
						else
							{
							$norm_dS = $norm_dSlist{$treeID1}{$treeID2} = sprintf ("%.3f", $dS / ((2 * $taxNum) - 6));
							}
					}

			last if ($compType eq "adjacent");
			}
		}
	return $dS, $norm_dS if (scalar(@treeID == 2));
	}

sub dotPlotter
	{
	my $rootingStatus = shift;
		print "Rooting status: $rootingStatus\n" if ($verbose);
	my $tree = shift;
		print "Tree description:\n\t$tree\n" if ($verbose);
	
	my (%ParentClade, %CladeDescrip, %CladeSupport, %open, %start, %stop, %present);
	my (@taxonList, @dotPlots, %plotCount, %taxonID, %dotPlot);
	
	my $totalClade = ($tree =~ tr/(//);
	my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
	my $OldCurrent;
	
	# Get list of taxa and assign IDs
		(my $taxonString = $tree) =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
			$taxonString =~ s/\(//g;
			$taxonString =~ s/\)//g;
		@taxonList = sort(split(/,/, $taxonString));
			for (my $i = 0; $i < scalar(@taxonList); $i++)
				{
				$taxonID{$taxonList[$i]} = $i;
				}

	# Parse tree description and derive dot plots		
		while ($position < length($tree))
			{
			my $offset = 1;
			if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$CladeID++;
				$ParentClade{$CladeID} = $CurrentClade;
				$CurrentClade = $CladeID;
				$open{$CurrentClade} = 1;
				$start{$CladeID} = $position;
				$CladeSupport{$CladeID} = 1;	# Support defaults to 1 for all clades; only reset as needed; NOTE: modification from standard module
				$dotPlot{$CladeID} = "." x scalar(@taxonList);	# Create blank dot plot for clade
				}
			elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				$stop{$OldCurrent} = $position;
				$CladeDescrip{$OldCurrent} = substr($tree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
				}
			elsif (substr($tree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or $position+$offset >= length($tree))
					{
					$offset++;
					}
				$position += $offset-1;
				}
			elsif (substr($tree, $position, 1) eq ",")
				{
				# Do nothing
				}
			elsif (substr($tree, $position, 1) eq "[")	# Skip over embedded comment
				{
				until (substr($tree, $position+$offset, 1) eq "]" or $position+$offset >= length($tree))
					{
					$offset++;
					}
				$position += $offset-1;
				}
			else	# Determine membership of clade
				{
				my $prevChar = substr($tree, $position-1, 1);
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or $position+$offset >= length($tree))
					{
					$offset++;
					}
				unless ($prevChar eq ")")	# Text represents a node label and not a taxon
					{
					my $taxon = substr($tree, $position, $offset);
					for (my $clade = 1; $clade <= $CladeID; $clade++)	# Assign taxon to any open clades
						{
						if ($open{$clade})
							{
							$present{substr($tree, $position, $offset)}{$clade} = 1;
							substr($dotPlot{$clade}, $taxonID{$taxon}, 1) = "*";
							}
						}
					}
				$position += $offset-1;
				}
			$position++;
			}

	# Check that dot plots are informative and invert as needed
		for (my $clade = 1; $clade <= $totalClade; $clade++)
			{
			print "Clade $clade:\n\t$CladeDescrip{$clade}\n" if ($verbose);
			my $dotPlot = $dotPlot{$clade};

			# Only store informative clades
				my $starCount = ($dotPlot =~ tr/\*//);
					next if ($starCount == 0 or $starCount == length($dotPlot));
					next if ($rootingStatus eq "unrooted" and ($starCount == 1 or $starCount == (length($dotPlot) - 1)));

			$dotPlot = invertPlot($dotPlot) if ($dotPlot =~ /^\*/ and ($rootingStatus eq "unrooted" or $rootingStatus eq "[&U]"));	# Invert plot to common frame of reference to account for unrooted trees

			$cladeWeight{$dotPlot} = $CladeSupport{$clade};	# NOTE: modification from standard module
			$plotCount{$dotPlot}++;
				print "\tDot plot: $dotPlot\n" if ($verbose);
				if ($plotCount{$dotPlot} == 1)	# Only store unique plots
					{
					push @dotPlots, $dotPlot;
					}
				else
					{
					print "\t\tNot unique ...\n" if ($verbose);
					}
			}
			
	return @dotPlots;
	}

sub invertPlot	# Change *s to .s (and vice versa) in dot plot
	{
	my $plot = shift;
		$plot =~ s/\./x/g;
		$plot =~ s/\*/\./g;
		$plot =~ s/x/\*/g;
		
	return $plot;
	}

sub pruneTree
	{
	my @treeList = @_;
	my (%taxonCount, @taxonList, @finalList);

	# Get taxon names on all trees
		print "Checking taxon names ...\n" if ($debug);	
		foreach my $treeID (@treeList)
			{
			my $tree = $treeDescription{$treeID};
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g unless ($keepSupport);
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g;
				$tree =~ s/\(//g;
				$tree =~ s/\)//g;
				print "\t$tree\n" if ($debug);

			foreach my $taxon (split (/,/, $tree))
				{
				$taxon =~ s/\s+/_/g;
				$taxonCount{$taxon}++;
				push @taxonList, $taxon if ($taxonCount{$taxon} == 1);
				}
			}
	
	# Process trees, keeping only those taxa on all trees
		print "Pruning taxa ...\n" if ($debug);	
		foreach my $treeID (@treeList)
			{
			my $tree = $treeDescription{$treeID};
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g unless ($keepSupport);
				print "\t$tree\n" if ($debug);
			undef @treeSpecies;
			foreach my $taxon (@taxonList)
				{
				if ($pruneType eq "common" and $taxonCount{$taxon} == scalar(@treeList))
					{
					push @treeSpecies, $taxon;
					}
				elsif ($pruneType eq "directed" and not defined $pruneTaxon{$taxon})
					{
					push @treeSpecies, $taxon;
					}
				elsif ($pruneType eq "both" and $taxonCount{$taxon} == scalar(@treeList) and not defined $pruneTaxon{$taxon})
					{
					push @treeSpecies, $taxon;
					}
				else	# Remove taxon along with any associated branch length
					{
					$tree =~ s/\($taxon(\:\d+\.?\d*e?\-?\d*)?\)/\(\)/g;	# Prefix = ( suffix = )
					$tree =~ s/\($taxon(\:\d+\.?\d*e?\-?\d*)?\,/\(\,/g;	# Prefix = ( suffix = ,
					$tree =~ s/\,$taxon(\:\d+\.?\d*e?\-?\d*)?\)/\,\)/g;	# Prefix = , suffix = )
					$tree =~ s/\,$taxon(\:\d+\.?\d*e?\-?\d*)?\,/\,\,/g;	# Prefix = , suffix = ,
					}
				}
			print "\t$tree\n" if ($debug);
			
			$prunedTree{$treeID} = cleanTree($tree);
			}
	}

sub cleanTree	# Subroutine to remove extraneous markers from tree description due to pruning taxa
	{
	my $dirtyTree = shift;
	print "\t\tCleaning tree ...\n" if ($debug);

	my $changeCount = 1;
	until ($changeCount == 0)	# Remove most empty clade markers from pruned ST description
		{
		$changeCount = $dirtyTree =~ s/,,/,/g;
		$changeCount += $dirtyTree =~ s/\(\)(:\d+\.?\d*e?\-?\d*)?//g;
		$changeCount += $dirtyTree =~ s/\(,\)(:\d+\.?\d*e?\-?\d*)?//g;
		$changeCount += $dirtyTree =~ s/\(,\(/\(\(/g;
		$changeCount += $dirtyTree =~ s/\),\)/\)\)/g;
		$changeCount += $dirtyTree =~ s/,\)/\)/g;
		$changeCount += $dirtyTree =~ s/\(,/\(/g;

		foreach my $species (@treeSpecies)	# Check for "clades" consisting of a single species: "(species:support)"
			{
			if ($keepSupport and $dirtyTree =~ /(\($species\:\d+\.?\d*e?\-?\d*\)\:\d+\.?\d*e?\-?\d*)/)
				{
				my $replaceText = $1;
					$replaceText =~ s/^\(//;
					$replaceText =~ s/\)\:\d+\.?\d*e?\-?\d*$//;
					
				$dirtyTree =~ s/\($species\:\d+\.?\d*e?\-?\d*\)\:\d+\.?\d*e?\-?\d*/$replaceText/;
				$changeCount++;
				}
			if ($dirtyTree =~ /\($species\)/)
				{
				$dirtyTree =~ s/\($species\)/$species/;
				$changeCount++;
				}
			}
		}
	print "\t\t\tDirty tree: $dirtyTree\n" if ($debug);

	# Fix for trees synonymized to a single taxon
		if (substr($dirtyTree, 0, 1) ne "(" and substr($dirtyTree, length($dirtyTree), 1) ne ")")
			{
			$dirtyTree = "(".$dirtyTree.")";
			}
			
	my $cleanTree = cladeClean($dirtyTree);
		$cleanTree =~ s/\:\d+\.?\d*e?\-?\d*$//;	# Strip trailing support value
	print "\t\t\tClean tree: $cleanTree\n" if ($debug);
	return $cleanTree;
	}

sub cladeClean	# Subroutine to remove "empty" clades from tree description
	{
	my $workTree = shift;
		print "\t\t\t\tWorking tree: $workTree\n" if ($debug);

	my ($position, $cladeID, $currentClade, $offset) = (0, 0, 0, 0);
	my $oldCurrent;
	my (%parentClade, %cladeDescrip, %cladeSupport, %start, %stop);
		$start{0} = 0;
		$stop{0}= length($workTree);
	my (%present, %deletedClade);
	my @delPositions;	

	# Get clade descriptions
		while ($position < length($workTree))
			{
			$offset = 1;
			if (substr($workTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeID++;
				$parentClade{$cladeID} = $currentClade;
				$currentClade = $cladeID;
				$start{$cladeID} = $position;
				$deletedClade{$cladeID} = 0;
				}
			elsif (substr($workTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$oldCurrent = $currentClade;
				$currentClade = $parentClade{$oldCurrent};
				$stop{$oldCurrent} = $position;

				# Get clade count to check for duplicate clades
					$cladeDescrip{$oldCurrent} = substr($workTree, $start{$oldCurrent}, $stop{$oldCurrent} - $start{$oldCurrent} + 1);
						$cladeDescrip{$oldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//g;
						$cladeDescrip{$oldCurrent} =~ s/\(//g;
						$cladeDescrip{$oldCurrent} =~ s/\)//g;
						
					$present{$cladeDescrip{$oldCurrent}}++;
					print "\t\t\t\t\tClade $oldCurrent: $cladeDescrip{$oldCurrent} ($present{$cladeDescrip{$oldCurrent}})\n" if ($debug);
					$deletedClade{$oldCurrent} = 1 if ($present{$cladeDescrip{$oldCurrent}} > 1);	# Clade is identical to one nested within it: delete
				}
			elsif (substr($workTree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or $position+$offset >= length($workTree))
					{
					$offset++;
					}
				if (substr($workTree, $position-1, 1) eq ")")	# Support value pertains to a clade
					{
					$cladeSupport{$oldCurrent} = substr($workTree, $position+1, $offset-1);
					}
				$position += $offset-1;
				}
			elsif (substr($workTree, $position, 1) eq ",")
				{
				}
			else	#Determine membership of clade
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq ",")
					{
					$offset++;
					}
				$position += $offset-1;
				}
			$position++;
			}
		
	# Reparse tree and print parentheses and clade supports unless clade is marked for deletion
		my $growTree;
		$position = $cladeID = 0;
		while ($position < length($workTree))
			{
			$offset = 1;
			if (substr($workTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeID++;
				$growTree .= "(" unless ($deletedClade{$cladeID});
				$currentClade = $cladeID;
				}
			elsif (substr($workTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$oldCurrent = $currentClade;
				unless ($deletedClade{$currentClade})
					{
					$growTree .= ")";
					$growTree .= ":$cladeSupport{$currentClade}" if ($keepSupport and defined $cladeSupport{$currentClade});
					}
				$currentClade = $parentClade{$oldCurrent};
				}
			elsif (substr($workTree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or $position+$offset >= length($workTree))
					{
					$offset++;
					}
				unless (substr($workTree, $position-1, 1) eq ")")	# Support value pertains to a terminal taxon
					{
					$growTree .= ":" . substr($workTree, $position+1, $offset-1) if ($keepSupport);
					}
				$position += $offset-1;
				}
			elsif (substr($workTree, $position, 1) eq ",")
				{
				$growTree .= ",";
				}
			else	#Determine membership of clade
				{
				$growTree .= substr($workTree, $position, 1);
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or substr($workTree, $position+$offset, 1) eq ":")
					{
					$growTree .= substr($workTree, $position+$offset, 1);
					$offset++;
					}
				$position += $offset-1;
				}
			$position++;
			}

	$growTree =~ s/\s//g;
	print "\t\tEmpty clean: $growTree\n" if ($debug);
	return $growTree;
	}

sub lnTransform	# Transform rates as ln(value + x) to correct for fact that absolute rates can differ by an order of magnitude or more between genes (differences become scaled); x is a tricky correction for possible zero rates involving adding the exponent to each value to maintain scale
	{
	my $testValue = shift;
	my $refValue = shift;
	
	# Get exponents; using floor (and not ceil) because ln values are negative
		my $testExp = floor(log($testValue) / log(10)) if ($testValue > 0);
		my $refExp = floor(log($refValue) / log(10)) if ($refValue > 0);
	
	if ($testValue > 0 and $refValue > 0)	# Add exponent of each value to itself
		{
		$testValue = log($testValue + 10**$testExp);
		$refValue = log($refValue + 10**$refExp);
		}
	elsif ($testValue == 0 and $refValue > 0)	# Add exponent of non-zero reference value to both values
		{
		$testValue = log($testValue + 10**$refExp);
		$refValue = log($refValue + 10**$refExp);
		}
	elsif ($testValue > 0 and $refValue == 0)	# Add exponent of non-zero test value to both values
		{
		$testValue = log($testValue + 10**$testExp);
		$refValue = log($refValue + 10**$testExp);
		}
	else	# Both are zero; leave as is
		{
		}

	return $testValue, $refValue;
	}

# Version history
#
#	v1.0 (May 18, 2006)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
